import numpy as np

def get_false_positive_rate_outcome(
    y_trues, y_preds, negative_value=0
):
    """Returns boolean outcome for the false positive rate. 1 if it is a false positive, 
    0 if it is a true negative, np.nan otherwhise.
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values.
    :param negative_value: value of the negative class.
    :returns: boolean outcome column values for the false positive rate.
    """
    fp = np.array(
        get_false_positives(y_trues, y_preds, negative_value=negative_value)
    ).astype(int)
    tn = np.array(
        get_true_negatives(y_trues, y_preds, negative_value=negative_value)
    ).astype(int)
    fp_outcome = np.full(fp.shape, np.nan)
    fp_outcome[fp == 1] = 1
    fp_outcome[tn == 1] = 0
    return fp_outcome


def get_false_negative_rate_outcome(
    y_trues, y_preds, positive_value=1
):
    """Returns boolean outcome for the false negative rate. 
    1 if it is a false negative, 0 if it is a true positive, np.nan otherwhise.
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values.
    :param positive_value: value of the positive class.
    :returns: boolean outcome column values for the false negative rate.
    """
    fn = np.array(
        get_false_negatives(y_trues, y_preds, positive_value=positive_value)
    ).astype(int)
    tp = np.array(
        get_true_positives(y_trues, y_preds, positive_value=positive_value)
    ).astype(int)
    fn_outcome = np.full(fn.shape, np.nan)
    fn_outcome[fn == 1] = 1
    fn_outcome[tp == 1] = 0
    return fn_outcome


def get_accuracy_outcome(y_trues, y_preds, negative_value=0, positive_value=1):
    """Returns boolean outcome for the accuracy rate. 1 if it is correct, 0 if it is incorrect.
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values.
    :param negative_value: value of the negative class.
    :param positive_value: value of the positive class.
    :returns: boolean outcome column values for the accuracy rate.
    """
    fn = np.array(
        get_false_negatives(y_trues, y_preds, positive_value=positive_value)
    ).astype(int)
    tp = np.array(
        get_true_positives(y_trues, y_preds, positive_value=positive_value)
    ).astype(int)
    fp = np.array(
        get_false_positives(y_trues, y_preds, negative_value=negative_value)
    ).astype(int)
    tn = np.array(
        get_true_negatives(y_trues, y_preds, negative_value=negative_value)
    ).astype(int)
    acc_outcome = np.full(y_trues.shape, np.nan)
    acc_outcome[(tp == 1) | (tn == 1)] = 1
    acc_outcome[(fp == 1) | (fn == 1)] = 0
    return acc_outcome


def get_true_positives(y_trues, y_preds, positive_value=1):
    """Returns true positives. True if it is a true positive, false otherwise.
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values
    :param positive_value: value of the positive class. 
    :returns: boolean outcome column values for the true positive rate.
    """
    return (y_trues == y_preds) & (y_trues == positive_value)


def get_true_negatives(y_trues, y_preds, negative_value=1):
    """Returns true negatives. True if it is a true negative, false otherwise.
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values.
    :param negative_value: value of the negative class.
    :returns: boolean outcome column values for the true negative rate.
    """
    return (y_trues == y_preds) & (y_trues == negative_value)


def get_false_positives(y_trues, y_preds, negative_value=1):
    """Returns false positives. True if it is a false positive, false otherwise
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values.
    :param negative_value: value of the negative class.
    :returns: boolean outcome column values for the false positive rate.
    """
    return (y_trues != y_preds) & (y_trues == negative_value)


def get_false_negatives(y_trues, y_preds, positive_value=1):
    """Returns false negatives. True if it is a false negative, false otherwise.
    :param y_trues: true values (e.g., df['y_true'])
    :param y_preds: predicated values.
    :param positive_value: value of the positive class.
    :returns: boolean outcome column values for the false negative rate.
    """
    return (y_trues != y_preds) & (y_trues == positive_value)



import collections

import numpy as np
import pandas as pd

def filterColumns(df_filter, cols):
    return df_filter[(df_filter[df_filter.columns[list(cols)]] > 0).all(1)]


def sum_values(_x):
    out = np.sum(_x, axis=0)
    return np.array(out).reshape(-1)


def setup_fptree(df, min_support, df_targets):
    num_itemsets = len(df.index)  # number of itemsets in the database

    is_sparse = False
    if hasattr(df, "sparse"):
        # DataFrame with SparseArray (pandas >= 0.24)
        if df.size == 0:
            itemsets = df.values
        else:
            itemsets = df.sparse.to_coo().tocsr()
            is_sparse = True
    else:
        # dense DataFrame
        itemsets = df.values

    # support of each individual item
    # if itemsets is sparse, np.sum returns an np.matrix of shape (1, N)
    item_support = np.array(np.sum(itemsets, axis=0) / float(num_itemsets))
    item_support = item_support.reshape(-1)

    items = np.nonzero(item_support >= min_support)[0]

    # Define ordering on items for inserting into FPTree
    indices = item_support[items].argsort()
    rank = {item: i for i, item in enumerate(items[indices])}

    if is_sparse:
        # Ensure that there are no zeros in sparse DataFrame
        itemsets.eliminate_zeros()

    # Building tree by inserting itemsets in sorted order
    # Heuristic for reducing tree size is inserting in order
    #   of most frequent to least frequent
    tree = FPTree(rank, trgs_root=np.zeros_like(df_targets[0]))
    for i in range(num_itemsets):
        if is_sparse:
            # itemsets has been converted to CSR format to speed-up the line
            # below.  It has 3 attributes:
            #  - itemsets.data contains non null values, shape(#nnz,)
            #  - itemsets.indices contains the column number of non null
            #    elements, shape(#nnz,)
            #  - itemsets.indptr[i] contains the offset in itemset.indices of
            #    the first non null element in row i, shape(1+#nrows,)
            nonnull = itemsets.indices[itemsets.indptr[i] : itemsets.indptr[i + 1]]
        else:
            nonnull = np.where(itemsets[i, :])[0]
        itemset = [item for item in nonnull if item in rank]
        itemset.sort(key=rank.get, reverse=True)

        tree.insert_itemset(itemset, trgs_i=df_targets[i].copy())

    return tree, rank


def generate_itemsets(
    generator, num_itemsets, colname_map, columns_accumulate=["tn", "fp", "fn", "tp"]
):
    itemsets = []
    supports = []
    c_dict = {}
    for i_c_dict in range(0, len(columns_accumulate)):
        c_dict[i_c_dict] = []
    for sup, iset, cf_final in generator:
        itemsets.append(frozenset(iset))
        supports.append(sup / num_itemsets)
        for i_c_dict in range(0, len(columns_accumulate)):
            c_dict[i_c_dict].append(cf_final[i_c_dict])

    res_dic = {"support": supports, "itemset": itemsets}

    res_dic.update({columns_accumulate[i]: c_dict[i] for i in c_dict})
    res_df = pd.DataFrame(res_dic)

    if colname_map is not None:
        res_df["itemset"] = res_df["itemset"].apply(
            lambda x: frozenset([colname_map[i] for i in x])
        )

    return res_df


def valid_input_check(df):
    if f"{type(df)}" == "<class 'pandas.core.frame.SparseDataFrame'>":
        msg = (
            "SparseDataFrame support has been deprecated in pandas 1.0,"
            " and is no longer supported in mlxtend. "
            " Please"
            " see the pandas migration guide at"
            " https://pandas.pydata.org/pandas-docs/"
            "stable/user_guide/sparse.html#sparse-data-structures"
            " for supporting sparse data in DataFrames."
        )
        raise TypeError(msg)

    if df.size == 0:
        return
    if hasattr(df, "sparse"):
        if not isinstance(df.columns[0], str) and df.columns[0] != 0:
            raise ValueError(
                "Due to current limitations in Pandas, "
                "if the sparse format has integer column names,"
                "names, please make sure they either start "
                "with `0` or cast them as string column names: "
                "`df.columns = [str(i) for i in df.columns`]."
            )

    # Fast path: if all columns are boolean, there is nothing to checks
    all_bools = df.dtypes.apply(pd.api.types.is_bool_dtype).all()
    if not all_bools:
        # Pandas is much slower than numpy, so use np.where on Numpy arrays
        if hasattr(df, "sparse"):
            if df.size == 0:
                values = df.values
            else:
                values = df.sparse.to_coo().tocoo().data
        else:
            values = df.values
        idxs = np.where((values != 1) & (values != 0))
        if len(idxs[0]) > 0:
            # idxs has 1 dimension with sparse data and 2 with dense data
            val = values[tuple(loc[0] for loc in idxs)]
            s = (
                "The allowed values for a DataFrame"
                " are True, False, 0, 1. Found value %s" % (val)
            )
            raise ValueError(s)


class FPTree:
    def __init__(self, rank=None, trgs_root=None):
        self.root = FPNode_CM(None, target_matrix=trgs_root)
        self.nodes = collections.defaultdict(list)
        self.cond_items = []
        self.rank = rank
        self.root_trgs_root = trgs_root

    def conditional_tree(self, cond_item, minsup):
        """
        Creates and returns the subtree of self conditioned on cond_item.
        Parameters
        ----------
        cond_item : int | str
            Item that the tree (self) will be conditioned on.
        minsup : int
            Minimum support threshold.
        Returns
        -------
        cond_tree : FPtree
        """
        # Find all path from root node to nodes for item
        branches = []
        count = collections.defaultdict(int)

        for node in self.nodes[cond_item]:
            branch = node.itempath_from_root()
            branches.append(branch)
            for item in branch:
                count[item] += node.count

        # Define new ordering or deep trees may have combinatorially explosion
        items = [item for item in count if count[item] >= minsup]
        items.sort(key=count.get)
        rank = {item: i for i, item in enumerate(items)}

        # Create conditional tree
        cond_tree = FPTree(rank, trgs_root=np.zeros_like(self.root_trgs_root))
        for idx, branch in enumerate(branches):
            branch = sorted(
                [i for i in branch if i in rank], key=rank.get, reverse=True
            )

            cond_tree.insert_itemset(
                branch,
                self.nodes[cond_item][idx].count,
                trgs_i=self.nodes[cond_item][idx].target_matrix.copy(),
            )
        cond_tree.cond_items = self.cond_items + [cond_item]

        return cond_tree

    def insert_itemset(self, itemset, count=1, trgs_i=None):
        """
        Inserts a list of items into the tree.
        Parameters
        ----------
        itemset : list
            Items that will be inserted into the tree.
        count : int
            The number of occurrences of the itemset.
        """

        self.root.count += count
        self.root.target_matrix += trgs_i

        if len(itemset) == 0:
            return

        # Follow existing path in tree as long as possible
        index = 0
        node = self.root
        for item in itemset:
            if item in node.children:
                child = node.children[item]
                child.count += count
                child.target_matrix += trgs_i.copy()
                node = child
                index += 1
            else:
                break

        # Insert any remaining items
        for item in itemset[index:]:
            child_node = FPNode_CM(item, count, node, target_matrix=trgs_i.copy())
            self.nodes[item].append(child_node)
            node = child_node

    def is_path(self):
        if len(self.root.children) > 1:
            return False
        for i in self.nodes:
            if len(self.nodes[i]) > 1 or len(self.nodes[i][0].children) > 1:
                return False
        return True

    def print_status(self, count, colnames):
        cond_items = [str(i) for i in self.cond_items]
        if colnames:
            cond_items = [str(colnames[i]) for i in self.cond_items]
        cond_items = ", ".join(cond_items)
        print(
            "\r%d itemset(s) from tree conditioned on items (%s)" % (count, cond_items),
            end="\n",
        )


class FPNode_CM:
    def __init__(self, item, count=0, parent=None, target_matrix=None):
        self.item = item
        self.count = count
        self.parent = parent
        self.children = collections.defaultdict(FPNode_CM)
        self.target_matrix = np.array(target_matrix)  # target_matrix.copy()

        if parent is not None:
            parent.children[item] = self

    def itempath_from_root(self):
        """Returns the top-down sequence of items from self to
        (but not including) the root node."""
        path = []
        if self.item is None:
            return path

        node = self.parent
        while node.item is not None:
            path.append(node.item)
            node = node.parent

        path.reverse()
        return path


# mlxtend Machine Learning Library Extensions
# Author: Steve Harenberg <harenbergsd@gmail.com>
#
# License: BSD 3 clause

import itertools
import math


def fpgrowth_cm(
    df,
    df_targets,
    min_support=0.5,
    use_colnames=True,
    max_len=None,
    verbose=0,
    columns_accumulate=["tn", "fp", "fn", "tp"],
):
    """Get frequent itemsets from a one-hot DataFrame
    Parameters
    -----------
    df : pandas DataFrame
      pandas DataFrame the encoded format. Also supports
      DataFrames with sparse data; for more info, please
      see (https://pandas.pydata.org/pandas-docs/stable/
           user_guide/sparse.html#sparse-data-structures)
      Please note that the old pandas SparseDataFrame format
      is no longer supported in mlxtend >= 0.17.2.
      The allowed values are either 0/1 or True/False.
      For example,
    ```
           Apple  Bananas   Beer  Chicken   Milk   Rice
        0   True    False   True     True  False   True
        1   True    False   True    False  False   True
        2   True    False   True    False  False  False
        3   True     True  False    False  False  False
        4  False    False   True     True   True   True
        5  False    False   True    False   True   True
        6  False    False   True    False   True  False
        7   True     True  False    False  False  False
    ```
    min_support : float (default: 0.5)
      A float between 0 and 1 for minimum support of the itemsets returned.
      The support is computed as the fraction
      transactions_where_item(s)_occur / total_transactions.
    use_colnames : bool (default: False)
      If true, uses the DataFrames' column names in the returned DataFrame
      instead of column indices.
    max_len : int (default: None)
      Maximum length of the itemsets generated. If `None` (default) all
      possible itemsets lengths are evaluated.
    verbose : int (default: 0)
      Shows the stages of conditional tree generation.
    Returns
    -----------
    pandas DataFrame with columns ['support', 'itemset'] of all itemsets
      that are >= `min_support` and < than `max_len`
      (if `max_len` is not None).
      Each itemset in the 'itemset' column is of type `frozenset`,
      which is a Python built-in type that behaves similarly to
      sets except that it is immutable
      (For more info, see
      https://docs.python.org/3.6/library/stdtypes.html#frozenset).
    Examples
    ----------
    For usage examples, please see
    http://rasbt.github.io/mlxtend/user_guide/frequent_patterns/fpgrowth/
    """
    valid_input_check(df)

    if min_support <= 0.0:
        raise ValueError(
            "`min_support` must be a positive "
            "number within the interval `(0, 1]`. "
            "Got %s." % min_support
        )
    df_targets = df_targets.values.copy()
    colname_map = None
    if use_colnames:
        colname_map = {idx: item for idx, item in enumerate(df.columns)}

    tree, _ = setup_fptree(df, min_support, df_targets)
    minsup = math.ceil(min_support * len(df.index))  # min support as count
    generator = fpg_step(tree, minsup, colname_map, max_len, verbose)

    return generate_itemsets(
        generator, len(df.index), colname_map, columns_accumulate=columns_accumulate
    )


def fpg_step(tree, minsup, colnames, max_len, verbose):
    """
    Performs a recursive step of the fpgrowth algorithm.
    Parameters
    ----------
    tree : FPTree
    minsup : int
    Yields
    ------
    lists of strings
        Set of items that has occurred in minsup itemsets.
    """
    count = 0
    items = tree.nodes.keys()

    if tree.is_path():
        # If the tree is a path, we can combinatorally generate all
        # remaining itemsets without generating additional conditional trees
        size_remain = len(items) + 1
        if max_len:
            size_remain = max_len - len(tree.cond_items) + 1
        for i in range(1, size_remain):
            for itemset in itertools.combinations(items, i):
                count += 1
                supports_t = {i: tree.nodes[i][0].count for i in itemset}
                from operator import itemgetter

                id_min, support = min(supports_t.items(), key=itemgetter(1))
                cf_y = tree.nodes[id_min][0].target_matrix
                yield support, tree.cond_items + list(itemset), cf_y
    elif not max_len or max_len > len(tree.cond_items):
        for item in items:
            count += 1
            support = sum([node.count for node in tree.nodes[item]])
            cf_y = sum([node.target_matrix for node in tree.nodes[item]])
            yield support, tree.cond_items + [item], cf_y

    if verbose:
        tree.print_status(count, colnames)

    # Generate conditional trees to generate frequent itemsets one item larger
    if not tree.is_path() and (not max_len or max_len > len(tree.cond_items)):
        for item in items:
            cond_tree = tree.conditional_tree(item, minsup)

            for sup, iset, cf_y in fpg_step(
                cond_tree, minsup, colnames, max_len, verbose
            ):
                yield sup, iset, cf_y



from itertools import chain, combinations
import math


def powerset(iterable):
    s = list(iterable)
    return [
        frozenset(i)
        for i in chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))
    ]


def compute_shapley_value_item(item, pattern, powerset_pattern, item_score):
    # Get all subsets of the pattern (powerset) where item is present
    subsets_item_i = [s for s in powerset_pattern if item.issubset(s)]

    # Compute the delta score for each subset containing the item
    deltas_item = {}

    for subset in subsets_item_i:
        subset_minus_item = subset - item

        # Delta(s) = v(s) - v(s \ {i})
        # The delta score is the score of the subset containing the item minus the score of the subset without the item
        deltas_item[subset] = (
            item_score[len(subset)][subset]
            - item_score[len(subset_minus_item)][subset_minus_item]
        )

    def weight_delta_score(s, n):
        """Weigh the delta scores
        s = |S| = len(subset) without the item
        n = |N| = len(pattern)
        """
        #  |S!|(|N| - |S| - 1)! / |N|!

        return (math.factorial(s) * math.factorial(n - s - 1)) / math.factorial(n)

    return sum(
        [  # We weight the delta score by |S!|(|N| - |S| - 1)! / |N|! where |S| = len(subset) without the item (len(k) - 1) and |N| = len(pattern)
            weight_delta_score(len(k) - 1, len(pattern)) * v
            for k, v in deltas_item.items()
        ]
    )


def compute_shapley_value(pattern: frozenset, item_score: dict):
    """Compute the Shapley value of a subset of items
    :param pattern: list of items
    :param item_score: dictionary of pattern scores: len(pattern) -> {pattern: score}
    """

    # Get all subsets of the pattern (powerset) - 2^N with N = len(pattern)
    powerset_pattern = powerset(pattern)

    # Initialize the Shapley value dictionary
    shapley_value = {}

    # For each item in the pattern
    for item_i in [frozenset([i]) for i in pattern]:
        # Compute the Shapley value for each item
        shapley_value[item_i] = compute_shapley_value_item(
            item_i, pattern, powerset_pattern, item_score
        )
    return shapley_value


#### Global Shapley Value


import numpy as np


def attribute(item):
    """Returns the attribute of an item.
    Args:
        item (str): item in the form of 'attribute=value'
    Returns:
        str: attribute of the item
    """
    return item.split("=")[0]


def plus(f1, f2):
    return frozenset(list(f1) + list(f2))


def weight_factor_global_shapley_value(lB, lA, lI, prod_mb):
    """Returns the weight factor for the global shapley value."""

    # \frac{|B|! (|A|-|B|-|I|)!} {|A|! \prod_{b \in B \cup attr(I)} m_b}
    import math

    return (math.factorial(lB) * math.factorial(lA - lB - lI)) / (
        math.factorial(lA) * (prod_mb)
    )


def global_itemset_divergence(I, scores_l, attributes, cardinality_attributes):
    """Returns the global divergence of an itemset.
    Args:
        I (set): itemset
        scores_l (dict): dictionary of len itemset, itemsets and their divergence
        attributes (list): list of attributes (from frequent items)
        card_map (dict): dictionary of attributes and their cardinality, i.e., how many values they can take (from frequent items)
    """
    # Get attributes different from the attribute of the item
    # These are our set B
    Bs = set(attributes) - {attribute(i) for i in I}

    # Get itemsets whose attributes are in B
    # These are our set I_B
    I_Bs = [
        pattern_i
        for len_patterns in scores_l
        for pattern_i in scores_l[len_patterns]
        if [item_i for item_i in pattern_i if attribute(item_i) not in Bs] == []
    ]

    # Initialize
    global_shapley_itemset = 0
    # J \in I_B
    for J in I_Bs:
        JI = plus(J, I)

        if len(JI) in scores_l and JI in scores_l[len(JI)]:
            # Get the attributes in J
            B = [attribute(i) for i in J]

            # Get the attributes in J \cup I
            attr_BI = B + [attribute(i) for i in I]

            prod_mb = np.prod([cardinality_attributes[i] for i in attr_BI])

            w = weight_factor_global_shapley_value(
                len(B), len(attributes), len(I), prod_mb
            )
            # Accumulate the global shapley value
            global_shapley_itemset = (
                global_shapley_itemset
                + (scores_l[len(JI)][JI] - scores_l[len(J)][J]) * w
            )

    return global_shapley_itemset



import numpy as np
import pandas as pd

from divexplorer.utils_FPgrowth import fpgrowth_cm


def get_df_one_hot_encoding(df_input):
    # One-hot encoding of the dataframe
    attributes = df_input.columns
    X_one_hot = pd.get_dummies(df_input, prefix_sep="=", columns=attributes)
    X_one_hot.reset_index(drop=True, inplace=True)
    return X_one_hot


def get_t_value_bayesian(positives, negatives, index_all_data):
    pos_plus_one = 1 + positives
    neg_plus_1 = 1 + negatives

    # Mean beta distribution
    mean_beta = pos_plus_one / (pos_plus_one + neg_plus_1)

    # Variance beta distribution
    variance_beta = (pos_plus_one * neg_plus_1) / (
        (pos_plus_one + neg_plus_1) ** 2 * (pos_plus_one + neg_plus_1 + 1)
    )

    mean_dataset, var_dataset = mean_beta[index_all_data], variance_beta[index_all_data]
    t_test_value = (abs(mean_beta - mean_dataset)) / (
        (variance_beta + var_dataset) ** 0.5
    )

    return t_test_value


def compute_welch_t_test(
    squared_values, support_count_values, mean_values, index_all_data
):
    variance_values = squared_values / support_count_values - mean_values**2
    mean_dataset = mean_values[index_all_data]
    variance_dataset = variance_values[index_all_data]
    support_count_dataset = support_count_values[index_all_data]

    # Welch's t-test
    t_test_welch = (abs(mean_values - mean_dataset)) / (
        (
            variance_values / support_count_values
            + variance_dataset / support_count_dataset
        )
        ** 0.5
    )
    return t_test_welch


class DivergenceExplorer:
    def __init__(
        self,
        df,
        is_one_hot_encoding=False,
    ):
        """
        :param df: pandas dataframe.  The columns that one wishes to analyze with divexplorer should have discrete values. 
        :param is_one_hot_encoding: boolean. If True, the dataframe attributes that one wishes to analyze are already one-hot encoded.
        """
        # df_discrete: pandas dataframe with discrete values
        self.df = df

        # is_one_hot_encoding: boolean, if True, the dataframe attributes are already one-hot encoded
        self.is_one_hot_encoding = is_one_hot_encoding

    def get_pattern_divergence(
        self,
        min_support: float,
        boolean_outcomes: list = None,
        quantitative_outcomes: list = None,
        attributes: list = None,
        FPM_algorithm="fpgrowth",
        show_coincise=True,
    ):
        """
        Computes the divergence of the specified outcomes.  One can specify two types of outcomes: boolean and quantitative.
        The difference lies mainly in the way in which the statistical significance is computed: in both cases, we use
        the Welch's t-test, but for boolean outcomes, we consider the outcomes as Bernoulli random variables. 
        One can specify multiple outcomes simultaneously, as a way to speed up the computation when multiple divergences are needed 
        (compared to computing them one by one).
        :param min_support: minimum support value for the pattern
        :param boolean_outcomes: list of boolean outcomes
        :param quantitative_outcomes: list of quantitative outcomes
        :param attributes: list of attributes to consider. If missing, all attributes except outcomes are considered. 
        :param FPM_algorithm: algorithm to use for frequent pattern mining
        :param show_coincise: if True, the output is more concise, returning only the average, the divergence and the t value
        """

        assert FPM_algorithm in [
            "fpgrowth",
            "apriori",
        ], f"{FPM_algorithm} algorithm is not handled. Qe integrate the DivExplorer computation in 'fpgrowth' and 'apriori' algorithms."

        # Sets the default values for lists. 
        quantitative_outcomes = quantitative_outcomes or []
        boolean_outcomes = boolean_outcomes or []
        attributes = attributes or []

        assert (
            len(boolean_outcomes) > 0 or len(quantitative_outcomes) > 0
        ), "At least one outcome must be specified."

        assert (
            len(boolean_outcomes) == 0 or len(quantitative_outcomes) == 0
        ), "Only one type of outcome must be specified."

        if attributes == []:
            # Get all attributes except outcomes
            attributes = [
                attr
                for attr in list(self.df.columns)
                if attr not in boolean_outcomes + quantitative_outcomes
            ]

        # Get only the attributes specified
        df_discrete = self.df[attributes]

        len_dataset = len(self.df)

        if self.is_one_hot_encoding == False:
            # If it is not already one-hot encoded, we one-hot encode it
            df_ohe = get_df_one_hot_encoding(df_discrete)

        if quantitative_outcomes:
            # If there are quantitative outcomes, we compute the squared outcome
            df_outcomes = self.df[quantitative_outcomes].copy()
            for outcome_name in quantitative_outcomes:
                # Compute the squared outcome - we will use it for the divergence computation

                df_outcomes.loc[:, f"{outcome_name}_SQUARED"] = (
                    np.array(self.df[outcome_name].values) ** 2
                )
        else:
            # We accumulate the outcomes
            df_outcomes = pd.DataFrame()

            for boolean_outcome in boolean_outcomes:
                positive_outcomes = (self.df[boolean_outcome] == 1).astype(int)
                negative_outcomes = (self.df[boolean_outcome] == 0).astype(int)
                bottom_outcomes = self.df[boolean_outcome].isna().astype(int)
                df_outcomes[f"{boolean_outcome}_positive"] = positive_outcomes
                df_outcomes[f"{boolean_outcome}_negative"] = negative_outcomes
                df_outcomes[f"{boolean_outcome}_bottom"] = bottom_outcomes

        if FPM_algorithm == "fpgrowth":
            df_divergence = fpgrowth_cm(
                df_ohe.copy(),  # Df with one-hot encoded attributes
                df_outcomes,  # Df with outcomes
                min_support=min_support,  # Minimum support
                columns_accumulate=list(df_outcomes.columns),  # Columns to accumulate
            )
        else:
            # We use the apriori algorithm
            if quantitative_outcomes:
                raise ValueError(
                    "The apriori implementation is available only for boolean outcomes."
                )
            df_with_outcomes = pd.concat([df_ohe, df_outcomes], axis=1)

            from divexplorer.utils_apriori import apriori_divergence

            df_divergence = apriori_divergence(
                df_ohe.copy(),
                df_with_outcomes,
                min_support=min_support,
            )

        all_dataset_row = {"support": 1, "itemset": frozenset()}

        cols_to_drop = []
        squared_cols_to_drop = []
        
        if boolean_outcomes:
            for boolean_outcome in boolean_outcomes:
                # The result is average when non considering the bottom values

                def compute_outcome(positive_outcomes, negative_outcomes):
                    return positive_outcomes / (positive_outcomes + negative_outcomes)

                positive_col_name = f"{boolean_outcome}_positive"
                negative_col_name = f"{boolean_outcome}_negative"
                bottom_col_name = f"{boolean_outcome}_bottom"

                df_divergence[boolean_outcome] = compute_outcome(
                    df_divergence[positive_col_name],
                    df_divergence[negative_col_name],
                )

                cols_to_drop.extend(
                    [positive_col_name, negative_col_name, bottom_col_name]
                )

                # Add the info of the all dataset row
                for column_name in [
                    positive_col_name,
                    negative_col_name,
                    bottom_col_name,
                ]:
                    all_dataset_row[column_name] = df_outcomes[column_name].sum()

                all_dataset_row[
                    f"{boolean_outcome}_div"
                ] = 0  # The divergence is 0 by definition
                all_dataset_row[
                    f"{boolean_outcome}_t"
                ] = 0  # The t value is 0 by definition

                # Compute the average of the all dataset row -- as above
                overall_average = compute_outcome(
                    all_dataset_row[positive_col_name],
                    all_dataset_row[negative_col_name],
                )

                # Add the average to the all dataset row
                all_dataset_row[boolean_outcome] = overall_average

                # Compute the divergence
                df_divergence[f"{boolean_outcome}_div"] = (
                    df_divergence[boolean_outcome] - overall_average
                )

                # Compute the t value

                # Get the positive and negative values
                pos, neg = (
                    df_divergence[positive_col_name].values,
                    df_divergence[negative_col_name].values,
                )

                # We append the pos and neg values of the all dataset row
                pos = np.concatenate([[all_dataset_row[positive_col_name]], pos])
                neg = np.concatenate([[all_dataset_row[negative_col_name]], neg])
                t = get_t_value_bayesian(pos, neg, 0)
                df_divergence[f"{boolean_outcome}_t"] = t[
                    1:
                ]  # We omit the all dataset row

        else:
            for quantitative_outcome in quantitative_outcomes:
                df_divergence[quantitative_outcome] = (
                    df_divergence[quantitative_outcome]
                    / (df_divergence["support"] * len_dataset).round()
                )

                quantitative_outcome_squared = f"{quantitative_outcome}_SQUARED"
                squared_cols_to_drop.append(quantitative_outcome_squared)

                # Add the info of the all dataset row
                for column_name in [
                    quantitative_outcome,
                    quantitative_outcome_squared,
                ]:
                    all_dataset_row[column_name] = df_outcomes[column_name].sum()

                all_dataset_row[
                    f"{quantitative_outcome}_div"
                ] = 0  # The divergence is 0 by definition
                all_dataset_row[
                    f"{quantitative_outcome}_t"
                ] = 0  # The t value is 0 by definition

                # Compute the average of the all dataset row -- as above
                overall_average = all_dataset_row[quantitative_outcome] / len_dataset

                # Add the average to the all dataset row
                all_dataset_row[quantitative_outcome] = overall_average

                # Compute the divergence
                df_divergence[f"{quantitative_outcome}_div"] = (
                    df_divergence[quantitative_outcome] - overall_average
                )

                # Compute the t value with Welch's t-test
                squared_values = df_divergence[quantitative_outcome_squared].values
                support_count_values = (
                    df_divergence["support"].values * len_dataset
                ).round()
                mean_values = df_divergence[quantitative_outcome].values

                # We append the pos and neg values of the all dataset row
                squared_values = np.concatenate(
                    [[all_dataset_row[quantitative_outcome_squared]], squared_values]
                )
                support_count_values = np.concatenate(
                    [[len_dataset], support_count_values]
                )
                mean_values = np.concatenate(
                    [[all_dataset_row[quantitative_outcome]], mean_values]
                )

                t = compute_welch_t_test(
                    squared_values, support_count_values, mean_values, 0
                )
                df_divergence[f"{quantitative_outcome}_t"] = t[
                    1:
                ]  # We omit the all dataset row

        # Add the all dataset row
        df_divergence.loc[
            len(df_divergence), all_dataset_row.keys()
        ] = all_dataset_row.values()

        df_divergence["length"] = df_divergence["itemset"].str.len()
        df_divergence["support_count"] = (
            df_divergence["support"] * len_dataset
        ).round()
        df_divergence = df_divergence.reset_index(drop=True)

        df_divergence.sort_values("support", ascending=False, inplace=True)
        df_divergence = df_divergence.reset_index(drop=True)

        if show_coincise:
            df_divergence = df_divergence.drop(columns=cols_to_drop)
        df_divergence = df_divergence.drop(columns=squared_cols_to_drop)
        
        return df_divergence



import pandas as pd
import numpy as np
from mlxtend.frequent_patterns import fpcommon as fpc
from mlxtend.frequent_patterns.apriori import (
    generate_new_combinations,
    generate_new_combinations_low_memory,
)


def apriori_divergence(
    df,
    df_true_pred,
    min_support=0.5,
    use_colnames=True,
    max_len=None,
    verbose=0,
    low_memory=False,
    cols_orderTP=["tn", "fp", "fn", "tp"],
    sortedV="support",
):
    """

    Returns
    -----------
    pandas DataFrame with columns ['support', 'itemset'] of all itemsets
      that are >= `min_support` and < than `max_len`
      (if `max_len` is not None).
      Each itemset in the 'itemset' column is of type `frozenset`,
      which is a Python built-in type that behaves similarly to
      sets except that it is immutable
      (For more info, see
      https://docs.python.org/3.6/library/stdtypes.html#frozenset).
    Examples
    -----------
    For usage examples, please see
    http://rasbt.github.io/mlxtend/user_guide/frequent_patterns/apriori/
    """

    def filterColumns(df_filter, cols):
        return df_filter[(df_filter[df_filter.columns[list(cols)]] > 0).all(1)]

    def sum_values(_x):
        out = np.sum(_x, axis=0)
        return np.array(out).reshape(-1)

    def _support(_x, _n_rows, _is_sparse):
        """DRY private method to calculate support as the
        row-wise sum of values / number of rows
        Parameters
        -----------
        _x : matrix of bools or binary
        _n_rows : numeric, number of rows in _x
        _is_sparse : bool True if _x is sparse
        Returns
        -----------
        np.array, shape = (n_rows, )
        Examples
        -----------
        For usage examples, please see
        http://rasbt.github.io/mlxtend/user_guide/frequent_patterns/apriori/
        """
        out = np.sum(_x, axis=0) / _n_rows
        return np.array(out).reshape(-1)

    if min_support <= 0.0:
        raise ValueError(
            "`min_support` must be a positive "
            "number within the interval `(0, 1]`. "
            "Got %s." % min_support
        )

    fpc.valid_input_check(df)

    if hasattr(df, "sparse"):
        # DataFrame with SparseArray (pandas >= 0.24)
        if df.size == 0:
            X = df.values
        else:
            X = df.sparse.to_coo().tocsc()
        is_sparse = True
    else:
        # dense DataFrame
        X = df.values
        is_sparse = False
    support = _support(X, X.shape[0], is_sparse)
    ary_col_idx = np.arange(X.shape[1])
    support_dict = {0: 1, 1: support[support >= min_support]}
    itemset_dict = {0: [()], 1: ary_col_idx[support >= min_support].reshape(-1, 1)}
    conf_metrics = {
        0: np.asarray([sum_values(df_true_pred[cols_orderTP])]),
        1: np.asarray(
            [
                sum_values(filterColumns(df_true_pred, item)[cols_orderTP])
                for item in itemset_dict[1]
            ]
        ),
    }
    max_itemset = 1
    rows_count = float(X.shape[0])

    all_ones = np.ones((int(rows_count), 1))

    while max_itemset and max_itemset < (max_len or float("inf")):
        next_max_itemset = max_itemset + 1

        # With exceptionally large datasets, the matrix operations can use a
        # substantial amount of memory. For low memory applications or large
        # datasets, set `low_memory=True` to use a slower but more memory-
        # efficient implementation.
        if low_memory:
            combin = generate_new_combinations_low_memory(
                itemset_dict[max_itemset], X, min_support, is_sparse
            )
            # slightly faster than creating an array from a list of tuples
            combin = np.fromiter(combin, dtype=int)
            combin = combin.reshape(-1, next_max_itemset + 1)

            if combin.size == 0:
                break
            if verbose:
                print(
                    "\rProcessing %d combinations | Sampling itemset size %d"
                    % (combin.size, next_max_itemset),
                    end="",
                )

            itemset_dict[next_max_itemset] = combin[:, 1:]
            support_dict[next_max_itemset] = combin[:, 0].astype(float) / rows_count
            max_itemset = next_max_itemset
            # TODO
        else:
            combin = generate_new_combinations(itemset_dict[max_itemset])
            combin = np.fromiter(combin, dtype=int)
            combin = combin.reshape(-1, next_max_itemset)

            if combin.size == 0:
                break
            if verbose:
                print(
                    "\rProcessing %d combinations | Sampling itemset size %d"
                    % (combin.size, next_max_itemset),
                    end="",
                )

            if is_sparse:
                _bools = X[:, combin[:, 0]] == all_ones
                for n in range(1, combin.shape[1]):
                    _bools = _bools & (X[:, combin[:, n]] == all_ones)
            else:
                _bools = np.all(X[:, combin], axis=2)
            support = _support(np.array(_bools), rows_count, is_sparse)
            _mask = (support >= min_support).reshape(-1)
            if any(_mask):
                itemset_dict[next_max_itemset] = np.array(combin[_mask])
                support_dict[next_max_itemset] = np.array(support[_mask])
                conf_metrics[next_max_itemset] = np.asarray(
                    [
                        sum_values(filterColumns(df_true_pred, itemset)[cols_orderTP])
                        for itemset in itemset_dict[next_max_itemset]
                    ]
                )
                max_itemset = next_max_itemset
            else:
                # Exit condition
                break

    all_res = []
    for k in sorted(itemset_dict):
        support = pd.Series(support_dict[k])
        itemsets = pd.Series([frozenset(i) for i in itemset_dict[k]], dtype="object")
        # conf_matrix_col=pd.Series(list(conf_metrics[k]))
        conf_metrics_cols = pd.DataFrame(list(conf_metrics[k]), columns=cols_orderTP)

        res = pd.concat((support, itemsets, conf_metrics_cols), axis=1)
        all_res.append(res)

    res_df = pd.concat(all_res)
    res_df.columns = ["support", "itemset"] + cols_orderTP

    if use_colnames:
        mapping = {idx: item for idx, item in enumerate(df.columns)}
        res_df["itemset"] = res_df["itemset"].apply(
            lambda x: frozenset([mapping[i] for i in x])
        )

    res_df["length"] = res_df["itemset"].str.len()
    res_df["support_count"] = np.sum(res_df[cols_orderTP], axis=1)

    res_df.sort_values(sortedV, ascending=False, inplace=True)
    res_df = res_df.reset_index(drop=True)

    if verbose:
        print()  # adds newline if verbose counter was used

    return res_df



# These fields are available in the patterns dataframe and we always want to keep them
BASE_COLUMNS = ["support", "itemset", "length", "support_count"]


def sort_pattern(x, abbreviations={}):
    # Sort a pattern and abbreviate by replacing the abbreviations. 
    x = list(x)
    x.sort()
    x = ", ".join(x)
    for k, v in abbreviations.items():
        x = x.replace(k, v)
    return x


def abbreviate_dictionary(pattern_value, abbreviations):
    # Shapley values (dict) as input
    return {
        frozenset([sort_pattern(k, abbreviations=abbreviations)]): v
        for k, v in pattern_value.items()
    }


class DivergencePatternProcessor:
    """Class to process patterns and compute Shapley values."""
    
    def __init__(self, patterns, metric):
        """
        :param patterns: dataframe patterns
        :param metric: name of the field (without final `_div`) used for divergence
        """
        # We only want to keep the columns that are relevant to the metric + the base columns
        self.patterns = patterns[
            BASE_COLUMNS + [metric, f"{metric}_div", f"{metric}_t"]
        ]
        self.metric = metric
        self.dict_len_pattern_divergence = (
            self.dict_len_pattern_divergence_representation()
        )

    def dict_len_pattern_divergence_representation(self):
        """
        Define an intermediate representation of the patterns dataframe in the form
        Len itemset -> {pattern: divergence}
        :return: Mapping from item length to item divergence.
        """
        patterns = self.patterns
        metric = self.metric + "_div"
        d = patterns[["itemset", metric]].set_index("itemset").to_dict("index")
        return {
            k: {k1: v[metric] for k1, v in d.items() if k == len(k1)}
            for k in range(0, max(patterns["length"] + 1))
        }

    def shapley_value(self, pattern=None, row_idx=None):
        """
        Compute the Shapley value of a pattern
        We can specify the pattern either directly by specifying the pattern (frozen set) or row_idx of the pattern in the patterns dataframe
        Args:
            pattern (frozen set): list of items - if None, row_idx must be provided
            row_idx (int): row index of the pattern in the patterns dataframe - if None, pattern must be provided
        Returns:
            (dict) Shapley value of the pattern - {item: shapley value} for each item in the pattern
        """
        assert (
            pattern is None or row_idx is None
        ), "Either pattern or row_idx must be provided"
        assert (
            pattern is not None or row_idx is not None
        ), "Either pattern or row_idx must be provided"

        if row_idx is not None:
            pattern = self.patterns.iloc[row_idx]["itemset"]
        from divexplorer.shapley_value import compute_shapley_value

        return compute_shapley_value(pattern, self.dict_len_pattern_divergence)

    def plot_shapley_value(
        self,
        pattern : frozenset =None,
        row_idx : int =None,
        shapley_values : dict =None,
        figsize : tuple=(4, 3),
        abbreviations : dict ={},
        sort_by_value: bool =True,
        height : float =0.5,
        linewidth : float=0.8,
        labelsize: int =10,
        title : str="",
        x_label="",
        name_fig : str=None,
        save_fig : bool=False,
        show_figure : bool=True,
    ):
        """
        Plot the Shapley value of a pattern.
        Specify either pattern or row_idx or shapley_value.
        :param pattern: list of items
        :param row_idx: row index of the pattern in the patterns dataframe
        :param shapley_values: dictionary of pattern scores: {pattern: score}
        :param figsize: figure size
        :param abbreviations: dictionary of abbreviations to replace in the patterns - for visualization purposes
        :param sort_by_value: sort the Shapley values by value
        :param height: height of the bars
        :param linewidth: width of the bar border
        :param labelsize: size of the labels
        :param title: title of the plot
        :param x_label: x label
        :param name_fig: name of the figure
        :param save_fig: save the figure
        :param show_figure: show the figure
        """

        assert (
            pattern is None or row_idx is None or shapley_values is None
        ), "Either pattern or row_idx or shapley_value must be provided"

        assert (
            pattern is not None or row_idx is not None or shapley_values is not None
        ), "Either pattern or row_idx or shapley_value must be provided"

        if pattern is not None or row_idx is not None:
            shapley_values = self.shapley_value(pattern=pattern, row_idx=row_idx)

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1, figsize=figsize, dpi=100)

        if abbreviations:
            shapley_values = abbreviate_dictionary(shapley_values, abbreviations)

        # Dictionary of item in string format and value
        shapleyv_plot = {str(",".join(list(k))): v for k, v in shapley_values.items()}

        if sort_by_value:
            shapleyv_plot = {
                k: v for k, v in sorted(shapleyv_plot.items(), key=lambda item: item[1])
            }

        ax.barh(
            range(len(shapleyv_plot)),
            shapleyv_plot.values(),
            height=height,
            align="center",
            color="#7CBACB",
            linewidth=linewidth,
            edgecolor="#0C4A5B",
        )

        ax.set_yticks(range(len(shapleyv_plot)), minor=False)
        ax.set_yticklabels(list(shapleyv_plot.keys()), minor=False)
        ax.tick_params(axis="y", labelsize=labelsize)
        ax.tick_params(axis="x", labelsize=labelsize)
        ax.set_title(title, fontsize=labelsize)

        ax.set_xlabel(x_label, size=labelsize)

        if save_fig:
            name_fig = "shapley_value.pdf" if name_fig is None else name_fig

            plt.savefig(
                f"{name_fig}",
                bbox_inches="tight",
                pad=0.05,
                facecolor="white",
                transparent=False,
            )
        if show_figure:
            plt.show()
            plt.close()

    def redundancy_pruning(self, th_redundancy):
        """
        Prune the patterns that are redundant with respect to the divergence
        :param th_redundancy: threshold for redundancy
        :returns: a Pandas dataframe containing patterns without redundancy

        Let I and  I - {item i} be two patterns   (for example,  {sex=Male, age=<25} and {sex=Male})
        If exist an item i such that it absolute marginal contribution is lower than a threshold epsilon,
        i.e. abs( divergence(I) - divergence(I - {item i}) <= epsilon
        We can prune I. The pattern 𝐼 - {item i} captures the divergence of pattern 𝐼, since the inclusion of the item i only slightly alters the divergence
        In the example, we would keep just sex=Male
        We proceed in this way for all the patterns.

        """
        patterns_divergence = self.dict_len_pattern_divergence
        redundants = []
        for patterns_len_p in patterns_divergence.values():
            for pattern in patterns_len_p:
                for item in pattern:
                    itemset_minus_item = pattern - frozenset([item])
                    v_itemset_minus_item = patterns_divergence[len(itemset_minus_item)][
                        itemset_minus_item
                    ]

                    v_pattern = patterns_divergence[len(pattern)][pattern]
                    if abs(v_pattern - v_itemset_minus_item) <= th_redundancy:
                        redundants.append(pattern)
        patterns_not_red = self.patterns.loc[
            self.patterns.itemset.isin(redundants) == False
        ]
        return patterns_not_red

    def get_patterns(self, th_redundancy=None, sort_by_divergence=True):
        """
        Return the patterns
        :param th_redundancy: threshold for redundancy - if None, no redundancy pruning
        :param sort_by_divergence: sort the patterns by divergence
        :returns: a Pandas dataframe containing the patterns and their divergence
        """
        if th_redundancy is None:
            patterns = self.patterns
        else:
            patterns = self.redundancy_pruning(th_redundancy)
        if sort_by_divergence:
            return patterns.sort_values(by=[f"{self.metric}_div"], ascending=False)
        else:
            return patterns

    def global_shapley_value(self):
        """Compute the Global Shapley value of the patterns
        The Global Shapley value is a generalization of the Shapley value to the entire set of all items.
        It captures the role of an item in giving rise to divergence jointly with other attributes.
        :returns: A dictionary associating each item to its Global Shapley value.
        """
        # Get 1-itemsets
        items = [item for item in self.dict_len_pattern_divergence[1].keys()]

        # Get attributes of 1-itemsets
        attributes = list(set([list(item)[0].split("=")[0] for item in items]))

        # Get cardinality of attributes

        from collections import Counter

        # Get the cardinality of each attribute - Number of items with that attribute
        cardinality_attributes = dict(
            Counter([list(item)[0].split("=")[0] for item in items])
        )

        global_shapley = {}
        for item in items:
            # Compute Global Shapley value of item
            from divexplorer.shapley_value import global_itemset_divergence

            global_shapley[item] = global_itemset_divergence(
                item,
                self.dict_len_pattern_divergence,
                attributes,
                cardinality_attributes,
            )

        return global_shapley
